//+------------------------------------------------------------------+
//|                                                TimeServerDST.mqh |
//|                                    Copyright (c) 2024, marketeer |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+

#include <MQL5Book/PRTF.mqh>
#include <MQL5Book/DateTime.mqh>

//+------------------------------------------------------------------+
//| Server time zone and DST current information                     |
//+------------------------------------------------------------------+
struct ServerTimeZone  // according to history analysis of week opening hours
{
   int offsetGMT;      // time zone offset in seconds against UTC/GMT for current week
   int offsetDST;      // DST correction in seconds (included in offsetGMT, as per MQL5)
   bool supportDST;    // DST changes are detected in the quotes
};

//+------------------------------------------------------------------+
//| Aux struct for week opening hours stats                          |
//+------------------------------------------------------------------+
struct TimeZoneChange
{
  datetime weekstart;
  int before;
  bool DSTb;
  int after;
  bool DSTa;
};

//+------------------------------------------------------------------+
//| Aux conversion from week opening hour to UTC                     |
//+------------------------------------------------------------------+
int HourToUTC(const int current, const bool dst = false)
{
  if(current == INT_MIN) return INT_MIN;
  
  int result = current + 2 + dst; // +2 to get UTC offset
  // if DST is enabled it's shown in offsetGMT according to MT5 rules
  // NB. when DST is enabled, time is shifted forward in the time zone,
  // but for outside it is looked like the time is shifted back
  // (everything in this time zone is happened earlier for others)
  // so to get standard time of the zone we need to add DST
  result %= 24;
     
  // time zones are in the range [UTC-12,UTC+12]
  if(result > 12) result = result - 24;
  return result;
}

//+------------------------------------------------------------------+
//| Analogue of TimeGMTOffset() function for trade server            |
//| Pseudo-code: TimeGMTOffset() = TimeGMT() - TimeLocal()           |
//+------------------------------------------------------------------+
int TimeServerGMTOffset()
{
  return (int)(TimeGMT() - TimeTradeServer());
}

//+------------------------------------------------------------------+
//| Estimate server time DST mode correction from H1 quotes history  |
//+------------------------------------------------------------------+
int TimeServerDaylightSavings(
  const uint lookupYears = 0,
  const string symbol = NULL)
{
  return TimeServerZone(lookupYears, symbol).offsetDST;
}

//+------------------------------------------------------------------+
//| Estimate server time zone offset from H1 quotes history          |
//+------------------------------------------------------------------+
int TimeServerGMTOffsetEmpiric(
  const uint lookupYears = 0,
  const string symbol = NULL)
{
  return TimeServerZone(lookupYears, symbol).offsetGMT;
}

//+------------------------------------------------------------------+
//| Estimate if server is DST-enabled from H1 quotes history         |
//+------------------------------------------------------------------+
bool TimeServerDaylightSavingsSupported(
  const uint lookupYears = 0,
  const string symbol = NULL)
{
  return TimeServerZone(lookupYears, symbol).supportDST;
}

#define WRAPAROUND(LIMIT, VALUE) ((VALUE + LIMIT) % LIMIT)
#define WRAP24(V) WRAPAROUND(24, V)

//+------------------------------------------------------------------+
//| Estimate server time zone and DST mode from H1 quotes history    |
//+------------------------------------------------------------------+
ServerTimeZone TimeServerZone(
  const uint lookupYears = 0, // by default, all available bars, 3 year looks enough
  const string symbol = NULL) // can use symbol in services
{
  const int year = 365 * 24 * 60 * 60;
  const int threshold = 52 / 4; // a quarter of weeks in a year
  const int hour = 3600;

  // result variables
  ServerTimeZone st = {};
  ServerTimeZone empty = {INT_MIN, INT_MIN, false};
  
  datetime array[];
  TimeZoneChange changes[];
  
  ResetLastError();
  
  if((lookupYears &&
     (CopyTime(symbol, PERIOD_H1, TimeCurrent() - lookupYears * year, TimeCurrent(), array)) > 0)
  || (CopyTime(symbol, PERIOD_H1, 0, Bars(symbol, PERIOD_H1), array)) > 0)
  {
     // approx. 6000 bars per year should be acquired here
     const int n = ArraySize(array);
     #ifdef PRINT_DST_DETAILS
     PrintFormat("Got %d H1 bars, ~%d days", n, n / 24);
     #endif

     int hours[24] = {};
     bool dst[24] = {};
     int current = INT_MIN, previous = INT_MIN;
     for(int i = 0; i < n; ++i)
     {
        const ENUM_DAY_OF_WEEK weekday = TimeDayOfWeek(array[i]);
        // skip all days except Sunday and Monday
        if(weekday > MONDAY) continue;
        // lets analyze the first H1 bar of the trading week
        // find out an hour for the first bar after weekend
        previous = current;
        current = _TimeHour();
        if(previous != current) // keep track of TZ changes
        {
          TimeZoneChange change[1] =
          {{
             array[i],
             HourToUTC(previous, previous == WRAP24(current - 1)),
             previous == WRAP24(current - 1),
             HourToUTC(current, WRAP24(previous - 1) == current),
             WRAP24(previous - 1) == current
          }};
          ArrayInsert(changes, change, ArraySize(changes));
        }
        
        // collect stats for opening hours
        hours[current]++;
        
        // now skip 2 days to check next week
        i += 48;
     }
     
     #ifdef PRINT_DST_DETAILS
     Print("Week opening hours stats:");
     ArrayPrint(hours);
     Print("Time Zone changes (UTC±X before/after weekstart):");
     ArrayPrint(changes);
     #endif
     
     if(hours[current] <= threshold)
     {
        #ifdef PRINT_DST_DETAILS
        Print("Extraordinary week detected, trying starting hour of the previous week: ", previous);
        #endif
        if(hours[previous] <= threshold)
        {
           // TODO: refine results using previous weeks
           #ifdef PRINT_DST_DETAILS
           Print("2 last weeks are extraordinary, giving up");
           #endif
           return empty;
        }
        else
        {
           current = previous;
        }
     }
     
     int tz = 0, tzdst = 0;
     for(int i = 0; i < 24; ++i)
     {
       if(hours[i] > threshold)
       {
         tz++;
         if(hours[WRAP24(i - 1)] > threshold)
         {
           dst[i] = true;
           tzdst++;
         }
       }
     }
     
     #ifdef PRINT_DST_DETAILS
     PrintFormat("%d different timezones detected in quotes, %d DST candidates", tz, tzdst);
     #endif
     
     bool DSTpossible = false;
     if(tzdst)
     {
       for(int i = 1; i < ArraySize(changes); i++)
       {
         if(changes[i - 1].DSTa && !changes[i].DSTa)
         {
           const int day = TimeDayOfYear(TimeCurrent());
           if(day >= TimeDayOfYear(changes[i - 1].weekstart) && day <= TimeDayOfYear(changes[i].weekstart))
           {
             DSTpossible = true;
             break;
           }
         }
       }
     }
     
     st.supportDST = tz > 1;  // DST is most likely supported (or broker changed time zone many times)
     bool DST = dst[current]; // DST is probably enabled for current week

     int utc = HourToUTC(current, DST);
     
     #ifdef PRINT_DST_DETAILS
     PrintFormat("Server time offset: UTC%+d %s", utc, DST ? "DST" : "STD");
     #endif
     if(!DST && DSTpossible)
     {
       #ifdef PRINT_DST_DETAILS
       PrintFormat("Possible time offset: UTC%+d DST", utc + DSTpossible);
       #endif
       if(TimeServerGMTOffset() == (utc + DSTpossible) * -hour)
       {
         utc += DSTpossible;
         DST = true;
       }
     }
     st.offsetGMT = -utc * hour;
     st.offsetDST = -DST * hour;

     return st;
  }
  else
  {
     Print("Can't read quotes, error: ", _LastError, " ", E2S(_LastError));
  }
  return empty;
}

#undef WRAP24
#undef WRAPAROUND

//+------------------------------------------------------------------+
/*
  example code:
  
    #define PRINT_DST_DETAILS
    #include "TimeServerDST.mqh"
    
    void OnStart()
    {
      PRTF(TimeServerDaylightSavings());
    }
  
  example output:
   
    Got 20023 H1 bars, ~834 days
    Week opening hours stats:
    30  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0 83 54  0
    Time Zone changes (UTC±X before/after weekstart):
                [weekstart]    [before] [DSTb] [after] [DSTa]
    [0] 2021.07.25 21:00:00 -2147483648  false      -1  false
    [1] 2021.11.07 22:00:00           0   true       0  false
    [2] 2022.03.13 21:00:00           0  false       0   true
    [3] 2022.11.06 22:00:00           0   true       0  false
    [4] 2023.03.12 21:00:00           0  false       0   true
    [5] 2023.11.05 22:00:00           0   true       0  false
    [6] 2024.03.11 00:00:00           0  false       2  false
    3 different timezones detected in quotes, 1 DST candidates
    Server time offset: UTC+2 STD
    Possible time offset: UTC+3 DST
    TimeServerDaylightSavings()=-3600 / ok

*/
//+------------------------------------------------------------------+
